import subprocess
from vassal.ssh import SSH
import os


class ArgsSCP(object):
    def __init__(self):
        self.r = False
        self.server = None
        self.username = False
        self.put = False
        self.remote_path = ""
        self.pkey = None
        self.file = ""

    def check_fields(self):
        for i in [self.server, self.username, self.remote_path, self.file]:
            if i == "":
                raise Exception("Scp syntax problem")


class Terminal(object):
    def __init__(self, commands, ssh_password=None, pkey_path=None, pkey_password=None):
        self.commands = commands
        self.ssh_password = ssh_password
        self.pkey_path = pkey_path
        self.pkey_password = pkey_password
        self.cwd = os.path.dirname(os.path.realpath(__file__))

    def truncate_command(self):
        cmd_truncks = []
        cur_truck = self.init_trunk()
        for cmd in self.commands:
            cmd = cmd.strip()
            if cmd.startswith("cd"):
                self.update_cwd(cmd)
                cur_truck.commands.append(cmd)

            elif cmd.startswith("ssh"):
                cmd_truncks.append(cur_truck)
                cur_truck = self.init_trunk(cmd)

            elif cmd.startswith("scp"):
                cmd_truncks.append(cur_truck)
                cur_truck = self.init_trunk(cmd)
                cmd_truncks.append(cur_truck)
                cur_truck = self.init_trunk()

            elif cmd.startswith("exit"):
                cur_truck.commands.append(cmd)
                cmd_truncks.append(cur_truck)
                cur_truck = self.init_trunk()
            else:
                cur_truck.commands.append(cmd)

        cmd_truncks.append(cur_truck)
        cmd_truncks = [trunk for trunk in cmd_truncks if trunk.commands]
        return cmd_truncks

    def init_trunk(self, first_cmd=None):
        if first_cmd is None:
            commands = []
        else:
            commands = [first_cmd]
        cur_trunk = TerminalTrunk(commands, self.cwd, ssh_password=self.ssh_password, pkey_path=self.pkey_path,
                                  pkey_password=self.pkey_password)
        if first_cmd is not None:
            if first_cmd.startswith("ssh"):
                cur_trunk.ssh = True

            elif first_cmd.startswith("scp"):
                cur_trunk.scp = True
        return cur_trunk

    def update_cwd(self, cd_command):
        path = cd_command.split(" ")[-1]
        self.cwd = os.path.join(self.cwd, path)

    def run(self):
        all_trunks = self.truncate_command()
        for trunk in all_trunks:
            trunk.run()


class TerminalTrunk(object):
    def __init__(self, commands, cwd, ssh_password=None, pkey_path=None, pkey_password=None, ssh=False, scp=False):
        self.commands = commands
        self.ssh_password = ssh_password
        self.pkey_path = pkey_path
        self.pkey_password = pkey_password
        self.cwd = cwd
        self.ssh = ssh
        self.scp = scp

    def _run_local(self):
        change_cwd = "cd " + self.cwd
        commands = ";".join([change_cwd] + self.commands)
        process = subprocess.Popen(commands, stdout=subprocess.PIPE, shell=True)
        proc_stdout = process.communicate()[0].strip()
        return proc_stdout.decode()

    def _parse_ssh(self, cmd):
        cmd_seg = cmd.split()
        serverATuser = [c for c in cmd_seg if "@" in c]
        assert len(serverATuser) == 1
        serverATuser = serverATuser[0]
        username = serverATuser.split("@")[0]
        server = serverATuser.split("@")[-1]
        return username, server

    def _parse_scp(self, cmd):
        cmd_seg = cmd.split()
        args = ArgsSCP()
        i = 0
        while i < len(cmd_seg):
            if cmd_seg[i] == "-r":
                args.r = True
            if cmd_seg[i] == "-i":
                args.pkey = cmd_seg[i + 1]
                i += 1
            if "@" in cmd_seg[i]:
                args.username = cmd_seg[i].split("@")[0]
                server_path = cmd_seg[i].split("@")[1]
                assert ":" in server_path
                args.server = server_path.split(":")[0]
                args.remote_path = server_path.split(":")[1]
            i += 1

        if "@" in cmd_seg[-1]:
            args.put = True
            args.file = cmd_seg[-2]
        elif "@" in cmd_seg[-2]:
            args.file = cmd_seg[-1]
        else:
            raise Exception("Scp syntax problem")
        args.file = os.path.join(self.cwd, args.file)
        return args

    def _init_ssh(self):
        username, server = self._parse_ssh(self.commands[0])
        client = SSH(server, username, password=self.ssh_password, pkey_path=self.pkey_path,
                     pkey_password=self.pkey_password)
        return client

    def _run_ssh(self):
        client = self._init_ssh()
        results = client.run(self.commands[1:])
        print(results)
        return results

    def _run_scp(self):
        assert len(self.commands) == 1
        args = self._parse_scp(self.commands[0])
        args.check_fields()
        client = SSH(args.server, args.username, password=self.ssh_password, pkey_path=args.pkey,
                     pkey_password=self.pkey_password)  # TODO support different keys
        if args.put:
            client.scp_upload(args.file, remote_path=args.remote_path, recursive=args.r)
        else:
            client.scp_download(args.remote_path, recursive=args.r, local_path=args.file)

    def run(self):
        if self.commands[0].startswith("ssh"):
            self._run_ssh()
        elif self.commands[0].startswith("scp"):
            self._run_scp()
        else:
            self._run_local()


from scp import SCPClient
import threading
from binascii import hexlify
import getpass
import os
import socket
import sys
import hashlib
from cryptography.fernet import Fernet
from paramiko.py3compat import input
import paramiko
import pickle


class SSH(object):
    def __init__(self, server, username, password=None, pkey_path=None, pkey_password=None):
        self.ssh = paramiko.SSHClient()
        self.ssh.load_system_host_keys()
        self.server = server

        self.username = username
        self.password = password
        self.pkey = self.load_key(pkey_path, pkey_password)
        self.port = 22
        if self.server.find(":") >= 0:
            self.server, portstr = self.server.split(":")
            self.port = int(portstr)

    def load_key(self, pkey_path, pkey_password):
        if pkey_path is not None:
            try:
                pkey = paramiko.RSAKey.from_private_key_file(pkey_path)
                return pkey
            except paramiko.ssh_exception.PasswordRequiredException:
                if pkey_password is None:
                    raise Exception("Public Private key file is encrypted: please add pkey_password to decrypt in SSH")
                try:
                    pkey = paramiko.RSAKey.from_private_key_file(pkey_path, password=pkey_password)
                except paramiko.ssh_exception.SSHException:
                    raise Exception("Wrong password for Public Private key file")
                return pkey
        return None

    def _input_auth(self):
        try:
            self.ssh.connect(hostname=self.server, username=self.username, pkey=self.pkey, password=self.password,
                             port=self.port)
            t = self.ssh.get_transport()
            chan = t.open_session()
            chan.get_pty()
            chan.invoke_shell()
            return t, chan
        except paramiko.ssh_exception.AuthenticationException:
            return None

    def run_exec(self, cmd, stderr=False):
        (stdin, stdout, stderr_out) = self.ssh.exec_command(cmd)
        if stderr:
            return stdout.read().decode(), stderr_out.read().decode()
        else:
            return stdout.read().decode()

    def _writeall(self, sock):
        while True:
            data = sock.recv(256)
            if not data:
                sys.stdout.write("\r\n*** EOF ***\r\n\r\n")
                sys.stdout.flush()
                break
            sys.stdout.write(data.decode())
            sys.stdout.flush()

    def run(self, cmds):
        self.authenticate()
        for i in range(len(cmds)):
            if not cmds[i].endswith("\n"):
                cmds[i] += "\n"
        cmds.append("exit\n")
        chan = self.chan
        sys.stdout.write("Line-buffered terminal emulation. Press F6 or ^Z to send EOF.\r\n\r\n")
        writer = threading.Thread(target=self._writeall, args=(chan,))
        writer.start()
        try:
            for i in cmds:
                chan.send(i)
        except (EOFError, KeyboardInterrupt):
            pass

    def _progress(self, filename, size, sent):
        sys.stdout.write("%s\'s progress: %.2f%%   \r" % (filename, float(sent) / float(size) * 100))

    def _load_auth(self, cred_fname):
        credentials = pickle.load(open(cred_fname, 'rb'))
        key = credentials['key']
        cipher_suite = Fernet(key)
        password = cipher_suite.decrypt(credentials['password'])
        mode = credentials['mode']
        keypass = credentials['keypass']

        return mode, password.decode(), keypass

    def _save_auth(self, cred_fname, mode, password="", keypass=""):
        if not os.path.exists("mapping"):
            os.mkdir("mapping")
        credentials = {}
        key = Fernet.generate_key()
        credentials['key'] = key
        cipher_suite = Fernet(key)
        credentials['password'] = cipher_suite.encrypt(password.encode())
        credentials['mode'] = mode
        credentials['keypass'] = keypass
        pickle.dump(credentials, open(cred_fname, 'wb'))

    def _rsa(self, t, username, path=None, password=None):
        default_path = os.path.join(os.environ["HOME"], ".ssh", "id_rsa")
        if path is None:
            path = input("RSA key [%s]: " % default_path)
        if len(path) == 0:
            path = default_path
        try:
            key = paramiko.RSAKey.from_private_key_file(path)
        except paramiko.PasswordRequiredException:
            if password is None:
                password = getpass.getpass("RSA key password: ")
            key = paramiko.RSAKey.from_private_key_file(path, password)
        t.auth_publickey(username, key)
        return password, path

    def _dss(self, t, username, path=None, password=None):
        default_path = os.path.join(os.environ["HOME"], ".ssh", "id_dsa")
        if path is None:
            path = input("DSS key [%s]: " % default_path)
        if len(path) == 0:
            path = default_path
        try:
            key = paramiko.DSSKey.from_private_key_file(path)
        except paramiko.PasswordRequiredException:
            if password is None:
                password = getpass.getpass("DSS key password: ")

            key = paramiko.DSSKey.from_private_key_file(path, password)
        t.auth_publickey(username, key)
        return password, path

    def _manual_auth(self, t, username, hostname, mode, password, keypass):
        default_auth = "p"
        path = None
        if mode is None:
            mode = input("%s: Auth by (p)assword, (r)sa key, or (d)ss key? [%s] " % (hostname, default_auth))
        if len(mode) == 0:
            mode = default_auth

        if mode == "r":
            password, path = self._rsa(t, username=username, path=keypass, password=password)
        elif mode == "d":
            password, path = self._dss(t, username=username, path=keypass, password=password)
        else:
            if password is None:
                password = getpass.getpass("Password for %s@%s: " % (username, hostname))
            t.auth_password(username, password)
        return mode, password, path

    def _agent_auth(self, transport, username):
        """
        Attempt to authenticate to the given transport using any of the private
        keys available from an SSH agent.
        """

        agent = paramiko.Agent()
        agent_keys = agent.get_keys()
        if len(agent_keys) == 0:
            return

        for key in agent_keys:
            print("Trying ssh-agent key %s" % hexlify(key.get_fingerprint()))
            try:
                transport.auth_publickey(username, key)
                print("... success!")
                return
            except paramiko.SSHException:
                print("... nope.")

    def _compute_key_path(self, server, username):
        m = hashlib.md5()
        m.update(server.encode())
        m.update(username.encode())
        md5sum = str(m.hexdigest())
        key_path = "mapping/{}.plk".format(md5sum)
        return key_path

    def _socket_server(self, hostname):
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((hostname, self.port))
        except Exception as e:
            print("*** Connect failed: " + str(e))
            sys.exit(1)
        return sock

    def scp_upload(self, file, remote_path="", recursive=False):
        self.authenticate()
        scp_client = SCPClient(self.transport, progress=self._progress)
        if remote_path != "":
            scp_client.put(file, recursive=recursive, remote_path=remote_path)
        else:
            scp_client.put(file, recursive=recursive)
        scp_client.close()

    def scp_download(self, file, local_path="", recursive=False):
        self.authenticate()
        scp_client = SCPClient(self.transport, progress=self._progress)
        scp_client.get(file, recursive=recursive, local_path=local_path)
        scp_client.close()

    def authenticate(self, force=False, save_cred=True):
        try:
            chan_t = self._input_auth()
            if chan_t is not None:
                self.transport, self.chan = chan_t
                return
        except:
            pass

        key_path = self._compute_key_path(self.server, self.username)

        if not force and os.path.exists(key_path):
            mode, password, keypass = self._load_auth(key_path)
        else:
            mode, password, keypass = None, None, None

        sock = self._socket_server(self.server)
        try:
            t = paramiko.Transport(sock)
            try:
                t.start_client()
            except paramiko.SSHException:
                print("*** SSH negotiation failed.")
                sys.exit(1)
            try:
                keys = paramiko.util.load_host_keys(os.path.expanduser("~/.ssh/known_hosts"))
            except IOError:
                try:
                    keys = paramiko.util.load_host_keys(os.path.expanduser("~/ssh/known_hosts"))
                except IOError:
                    print("*** Unable to open host keys file")
                    keys = {}

            # check server's host key -- this is important.
            key = t.get_remote_server_key()
            if self.server not in keys:
                print("*** WARNING: Unknown host key!")
            elif key.get_name() not in keys[self.server]:
                print("*** WARNING: Unknown host key!")
            elif keys[self.server][key.get_name()] != key:
                print("*** WARNING: Host key has changed!!!")
                sys.exit(1)
            else:
                print("*** Host key OK.")
            # get username

            self._agent_auth(t, self.username)
            if not t.is_authenticated():
                mode, password, path = self._manual_auth(t, self.username, self.server, mode, password, keypass)
                if save_cred:
                    self._save_auth(cred_fname=key_path, mode=mode, password=password, keypass=path)
            if not t.is_authenticated():
                print("*** Authentication failed. :(")
                t.close()
                sys.exit(1)

            chan = t.open_session()
            chan.get_pty()
            chan.invoke_shell()
            self.chan = chan
            self.transport = t

        except Exception as e:
            print("*** Caught exception: " + str(e.__class__) + ": " + str(e))
            try:
                t.close()
            except:
                pass
            sys.exit(1)


import time


class Scheduler(object):
    def __init__(self, terminal, time_to_run=None, day=0, hour=0, min=0, sec=0):
        self.terminal = terminal
        self.day = day
        self.hour = hour
        self.min = min
        self.sec = sec
        self.time_to_run = time_to_run
        self.total_time = max(0.1, self.sec + 60 * self.min + 60 * 60 * self.hour + 60 * 60 * 24 * self.day)

    def run(self):
        if self.time_to_run is None:
            while True:
                self.terminal.run()
                time.sleep(self.total_time)
        else:
            for _ in range(self.time_to_run):
                self.terminal.run()
                time.sleep(self.total_time)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="vassal",
    version="0.0.2",
    author="Shawn Shan",
    author_email="shawnshan1825@gmail.com",
    description="The automate terminal",
    long_description=long_description,
    url="https://github.com/Shawn-Shan/vassal",
    packages=setuptools.find_packages(),
    install_requires=[
        'scp>=0.13.0',
        'cryptography>=2.4.1',
        'paramiko>=2.4.2',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
